// crunch.cc  Process samples.txt to stdout.

#include <cstdio>
#include <cstdlib>
#include <cmath>
#include <cassert>
#include <cstring>
#include <cinttypes>
#include "bench_constants.h"

#include <iostream>
#include <fstream>
#include <vector>
#include <string>
#include <algorithm>


const uint8_t maxCXXopt    =  4;
const uint8_t numOpt       =  7;
const uint8_t numCVariant  =  3;
const uint8_t numVariant   = 12;
const uint8_t numRotation  =  9;
static_assert (numOpt * numRotation <= UINT8_MAX, "numRun overflow");
const uint8_t numRun       = numOpt * numRotation;

const std::vector<std::string> opts = {
  "--O0", "--O1", "--O2", "--O3", "--O3 --flto",
  "--O3 --fmudflap --fmudflapir", "--O3 --fmudflap"
};

const std::vector<std::string> variants = {
  "C pointer", "C array[i]", "C array[x[i]]", "C++ pointer", "C++ array[i]",
  "C++ array[x[i]]", "C++ vector std::fill", "C++ vector iterator",
  "C++ vector[i]", "C++ vector.at(i)", "C++ vector[x[i]]",
  "C++ vector.at(x[i])"
};

// tinv((1+0.95^(1/866))/2,62)
const double k = 4.31100418960161135828;


// All values stored in nsec

// This structure is scope-agnostic (could be a series, could be a cell).
struct DataGroup {
  std::vector<uintmax_t> data;
  double mean;
  double std;
  double DurbinWatson;
  double normalEstimatedStandardDeviationOfMean;

  DataGroup() { mean = std = DurbinWatson
      = normalEstimatedStandardDeviationOfMean = 0.0; }
};

struct Cell {
  DataGroup cell;
  std::vector<DataGroup> serieses;
  double standardDeviationOfSeriesMeans;
  double estimatedStandardDeviationOfMeanOfSeriesMeans;
  double meanSeriesDurbinWatson;
  double expandedUncertainty;

  Cell() { standardDeviationOfSeriesMeans
      = estimatedStandardDeviationOfMeanOfSeriesMeans
      = meanSeriesDurbinWatson
      = expandedUncertainty
      = 0.0; }
};

typedef std::vector<Cell> Row;


// Fill in stats for a DataGroup.
void calcStats (DataGroup &d) {
  assert (d.data.size() > 1);
  d.mean = static_cast<double> (std::accumulate (d.data.begin(),
						 d.data.end(),
						 UINTMAX_C(0)))
    / d.data.size();
  double e2sum = 0.0, DWtopsum = 0.0, prev_e = 0.0;
  for (std::vector<uintmax_t>::iterator it = d.data.begin();
       it != d.data.end(); ++it) {
    double e = *it - d.mean;
    e2sum += e*e;
    if (it != d.data.begin()) {
      double x = e - prev_e;
      DWtopsum += x*x;
    }
    prev_e = e;
  }
  d.std = sqrt (e2sum / (d.data.size() - 1));
  d.normalEstimatedStandardDeviationOfMean =
    sqrt (e2sum / (d.data.size() * (d.data.size() - 1)));
  d.DurbinWatson = DWtopsum / e2sum;
  assert (d.DurbinWatson >= 0 && d.DurbinWatson <= 4);
}


// Fill in stats for a Cell and every DataGroup therein.
void calcStats (Cell &c) {
  assert (c.serieses.size() == numRun);
  assert (c.cell.data.size() == c.serieses.size() * repeats);

  calcStats (c.cell);
  std::for_each (c.serieses.begin(), c.serieses.end(),
		 (void(*)(DataGroup&))calcStats);

  // Calculate cell-level stats.
  double e2sum = 0, dsum = 0;
  uintmax_t count = 0;
  for (std::vector<DataGroup>::const_iterator it = c.serieses.begin();
       it != c.serieses.end(); ++it) {
    double e = it->mean - c.cell.mean;
    e2sum += e*e;
    count += it->data.size();
    dsum += it->DurbinWatson;
  }
  assert (count == c.cell.data.size());
  c.standardDeviationOfSeriesMeans = sqrt (e2sum / (numRun - 1));
  c.estimatedStandardDeviationOfMeanOfSeriesMeans
    = sqrt (e2sum / (numRun * (numRun - 1)));
  c.meanSeriesDurbinWatson = dsum / numRun;
  c.expandedUncertainty = k * c.estimatedStandardDeviationOfMeanOfSeriesMeans;
}


// Fill in stats for every Cell in a Row.
void calcStats (Row &r) {
  std::for_each (r.begin(), r.end(), (void(*)(Cell&))calcStats);
}


// Format a number < 1e6 with thin spaces per NIST SP 811.
#if 0
// Zero-decimals version
void fmt_usec (double nsec_in, std::string &out) {
  char buf[80];
  sprintf (buf, "%.0f", nsec_in / 1000.0);
  size_t l = strlen (buf);
  switch (l) {
  case 1:
  case 2:
  case 3:
  case 4:
    out = buf;
    break;
  case 5:
    out = buf[0];
    out += buf[1];
    out += "\\,";
    out += buf+2;
    break;
  case 6:
    out = buf[0];
    out += buf[1];
    out += buf[2];
    out += "\\,";
    out += buf+3;
    break;
  default:
    assert (false);
  }
}
#else
// 1-decimal version
void fmt_usec (double nsec_in, std::string &out) {
  char buf[80];
  sprintf (buf, "%.1f", nsec_in / 1000.0);
  size_t l = strlen(buf);
  if (l < 7)
    out = buf;
  else if (l == 7) {
    out = buf[0];
    out += buf[1];
    out += "\\,";
    out += buf+2;
  } else if (l == 8) {
    out = buf[0];
    out += buf[1];
    out += buf[2];
    out += "\\,";
    out += buf+3;
  } else
    assert (false);
}
#endif


char const header[] = "\
\\documentclass[10pt]{article}\n\
\\usepackage[pdftex]{graphicx}\n\
\\usepackage{lscape}\n\
\\pagestyle{empty}\n\
\\oddsidemargin      0 in\n\
\\evensidemargin     0 in\n\
\\topmargin          0 in\n\
\\headheight         0 in\n\
\\headsep            0 in\n\
\\textheight         9 in\n\
\\textwidth        6.5 in\n\
\\parindent          0 pt\n\
\\parskip            6 pt plus 6 pt\n\
\\partopsep         -6 pt\n\
\\pdfpageheight     11 in\n\
\\pdfpagewidth     8.5 in\n\
\\pdfcompresslevel   9\n\
\\tabcolsep          2 pt\n\
\\newcommand{\\badDW}{\\small}\n\
\\begin{document}\n";


// Functions to pass to printRow/printTable.
void cellMeanWithUncertainty (const Cell &c) {
  std::string fmtMean;
  fmt_usec (c.cell.mean, fmtMean);
  if (c.cell.DurbinWatson < 1)
    printf ("& \\emph{\\badDW %s $\\pm$ %0.1f} ", fmtMean.c_str(), c.expandedUncertainty/1000);
  else
    printf ("& %s $\\pm$ %0.1f ", fmtMean.c_str(), c.expandedUncertainty/1000);
}
void cellStd (const Cell &c) {
  std::string formatted;
  fmt_usec (c.cell.std, formatted);
  if (c.cell.DurbinWatson < 1)
    printf ("& \\multicolumn{1}{r}{\\emph{\\badDW %s}} ", formatted.c_str());
  else
    printf ("& \\multicolumn{1}{r}{%s} ", formatted.c_str());
}
void standardDeviationOfMeanOfSeriesMeans (const Cell &c) {
#if 0
  std::string formatted;
  fmt_usec (c.estimatedStandardDeviationOfMeanOfSeriesMeans, formatted);
  printf ("& %s ", formatted.c_str());
#else
  double usec = c.estimatedStandardDeviationOfMeanOfSeriesMeans / 1000.0;
  // 17 significant digits
  printf ("& %.*f ", 17-((int)floor(log10(usec))+1), usec);
#endif
}
void cellDurbinWatson (const Cell &c) {
  printf ("& %.1f ", c.cell.DurbinWatson);
}
void meanSeriesDurbinWatson (const Cell &c) {
  printf ("& %.1f ", c.meanSeriesDurbinWatson);
}
void cellUncertaintyRatio (const Cell &c) {
  printf ("& %.1f ",
    c.estimatedStandardDeviationOfMeanOfSeriesMeans /
    c.cell.normalEstimatedStandardDeviationOfMean
  );
}


// Print a row with content determined by the fuction parameter.
void printRow (const Row &r, void (*printer)(const Cell &),
	       unsigned firstCol, unsigned numCol) {
  // It was simpler before tables had to be split...
  // std::for_each (r.begin(), r.end(), printer);
  const unsigned end (std::min (firstCol+numCol, (unsigned)r.size()));
  for (unsigned i=firstCol; i<end; ++i)
    (*printer) (r.at(i));
  printf ("\\\\\n");
}


void printTableHeader (unsigned firstCol, unsigned numCol, char align) {
  printf ("\\begin{tabular}{c*{%u}{%c}}\n", numCol, align);
  for (unsigned i=firstCol; i<firstCol+numCol; ++i)
    printf ("& \\multicolumn{1}{c}{\\rotatebox{90}{%s}} ",
	    variants.at(i).c_str());
  printf ("\\\\\n");
}


// Print a table with content determined by the fuction parameter.
void printTable (const std::vector<Row> &table,
		 void (*printer)(const Cell &)) {
  printTableHeader (0, numVariant, 'r');
  for (uint8_t row=0; row<numOpt; ++row) {
    printf ("%s ", opts.at(row).c_str());
    printRow (table.at(row), printer, 0, numVariant);
  }
  printf ("\\end{tabular}\n\n");
}


// Split up table because page is too narrow, with align c.
void printSplitTable (const std::vector<Row> &table,
		      void (*printer)(const Cell &)) {
  static_assert (numVariant % 2 == 0, "fragile code in printSplitTable needs fixing");
  printTableHeader (0, numVariant/2, 'c');
  for (uint8_t row=0; row<numOpt; ++row) {
    printf ("%s ", opts.at(row).c_str());
    printRow (table.at(row), printer, 0, numVariant/2);
  }
  printf ("\\end{tabular}\n\n");
  printTableHeader (numVariant/2, numVariant/2, 'c');
  for (uint8_t row=0; row<numOpt; ++row) {
    // Omit blank rows
    if (table.at(row).size() > numVariant/2) {
      printf ("%s ", opts.at(row).c_str());
      printRow (table.at(row), printer, numVariant/2, numVariant/2);
    }
  }
  printf ("\\end{tabular}\n\n");
}


void printRatiosTable (const std::vector<Row> &smallTable,
		       const std::vector<Row> &bigTable) {
  printTableHeader (0, numVariant, 'r');
  for (uint8_t row=0; row<=maxCXXopt; ++row) {
    printf ("%s ", opts.at(row).c_str());
    for (uint8_t col=0; col<numVariant; ++col)
      printf ("& %.1f ", bigTable.at(row).at(col).cell.mean /
	      smallTable.at(row).at(col).cell.mean);
    printf ("\\\\\n");
  }
  for (uint8_t row=maxCXXopt+1; row<numOpt; ++row) {
    printf ("%s ", opts.at(row).c_str());
    for (uint8_t col=0; col<numCVariant; ++col)
      printf ("& %.1f ", bigTable.at(row).at(col).cell.mean /
	      smallTable.at(row).at(col).cell.mean);
    printf ("\\\\\n");
  }
  printf ("\\end{tabular}\n");
}


// Oversized spreadsheet, not for publication.
void hugeTable (const std::vector<Row> &table) {
  const float hugeWidth = 25, hugeHeight = 77;
  printf ("\\clearpage \\pdfpagewidth=%fin \\pdfpageheight=%fin \\enlargethispage{%fin}\n",
	  hugeWidth, hugeHeight, hugeHeight-11);
  printf ("\\begin{tabular}{c*{%u}{|rrrrr}}\n", numVariant);
  for (std::vector<std::string>::const_iterator it = variants.begin();
       it != variants.end(); ++it)
    printf ("& \\multicolumn{5}{|c}{\\rotatebox{90}{%s}} ", it->c_str());
  printf ("\\\\\n");
  for (uint8_t i=0; i<numVariant; ++i)
    printf ("& \\multicolumn{1}{|c}{min} & \\multicolumn{1}{c}{$\\mu$} & \\multicolumn{1}{c}{max} & \\multicolumn{1}{c}{s} & \\multicolumn{1}{c}{d} ");
  printf ("\\\\ \\hline\n");
  for (uint8_t rowNum=0; rowNum<numOpt; ++rowNum) {
    const Row &row (table.at(rowNum));
    for (uint8_t run=0; run<numRun; ++run) {
      if (run == 0)
        printf ("%s ", opts.at(rowNum).c_str());
      for (Row::const_iterator it = row.begin(); it != row.end(); ++it) {
        assert (it->serieses.size() == numRun);
        const DataGroup &series (it->serieses.at(run));
        printf ("& %.0f & %.0f & %.0f & %.0f & %.1f ",
	  *std::min_element(series.data.begin(),series.data.end()) / 1000.0,
	  series.mean                                              / 1000.0,
	  *std::max_element(series.data.begin(),series.data.end()) / 1000.0,
	  series.std                                               / 1000.0,
	  series.DurbinWatson);
      }
      if (run == numRun-1)
        printf ("\\\\ \\hline\n");
      else
        printf ("\\\\\n");
    }
  }
  printf ("\\end{tabular}\n\n");
}


// Scan and sort data.
void readData (std::vector<Row> &smallTable, std::vector<Row> &bigTable) {
  std::string buf;
  std::ifstream fp;
  fp.open (samples_fname, std::ifstream::in);
  getline (fp, buf);
  while (fp.good()) {
    unsigned optimization, variant;
    if (sscanf (buf.c_str(), "begin %u %u", &optimization, &variant) != 2) {
      fprintf (stderr, "Format error on begin line\n");
      exit (EXIT_FAILURE);
    }
    Cell &smallCell (smallTable.at(optimization).at(variant));
    Cell &bigCell (bigTable.at(optimization).at(variant));
    DataGroup bigSeries, smallSeries;
    for (uintmax_t samp=0; samp<repeats; ++samp) {
      getline (fp, buf);
      assert (fp.good());
      uintmax_t small, large;
      if (sscanf (buf.c_str(), "%" SCNuMAX " %" SCNuMAX, &small, &large) != 2) {
        fprintf (stderr, "Format error on sample line\n");
        exit (EXIT_FAILURE);
      }
      smallSeries.data.push_back (small);
      bigSeries.data.push_back (large);
    }
    smallCell.serieses.push_back (smallSeries);
    bigCell.serieses.push_back (bigSeries);
    smallCell.cell.data.insert (smallCell.cell.data.end(),
				smallSeries.data.begin(), smallSeries.data.end());
    bigCell.cell.data.insert (bigCell.cell.data.end(),
			      bigSeries.data.begin(), bigSeries.data.end());
    getline (fp, buf);
  }
  fp.close();
}


void usage (char const * const a0) __attribute__ ((noreturn));
void usage (char const * const a0) {
  fprintf (stderr, "Usage: %s                (to create crunch.tex)\n       %s row col [run]  (to dump largeTable cell or series data)\n",
	   a0, a0);
  exit (EXIT_FAILURE);
}


// Dump statistics to stderr, unformatted data (in ns) to stdout.
void dumpDataGroup (const DataGroup &d) {
  fprintf (stderr, "min mean max s d = %.0f %.0f %.0f %.0f %.1f\n",
    *std::min_element(d.data.begin(),d.data.end()) / 1000.0,
    d.mean                                         / 1000.0,
    *std::max_element(d.data.begin(),d.data.end()) / 1000.0,
    d.std                                          / 1000.0,
    d.DurbinWatson);
  for (std::vector<uintmax_t>::const_iterator it = d.data.begin();
       it != d.data.end(); ++it)
    printf ("%" PRIuMAX "\n", *it);
}


int main (int argc, char **argv) {
  bool dump = false, dumpEntireCell = true;
  unsigned dumpRow, dumpCol, dumpRun;

  // Parse args.
  switch (argc) {
  case 1:
    break;
  case 4:
    dumpEntireCell = false;
    if (sscanf (argv[3], "%u", &dumpRun) != 1)
      usage (argv[0]);
    if (dumpRun >= numRun) {
      fprintf (stderr, "run must be < %u\n", numRun);
      usage (argv[0]);
    }
  case 3:
    dump = true;
    if (sscanf (argv[1], "%u", &dumpRow) != 1)
      usage (argv[0]);
    if (sscanf (argv[2], "%u", &dumpCol) != 1)
      usage (argv[0]);
    if (dumpRow >= numOpt) {
      fprintf (stderr, "row must be < %u\n", numOpt);
      usage (argv[0]);
    }
    if (dumpCol >= numVariant) {
      fprintf (stderr, "col must be < %u\n", numVariant);
      usage (argv[0]);
    }
    if (dumpRow > maxCXXopt && dumpCol >= numCVariant) {
      fprintf (stderr, "No mudflap data for C++\n");
      usage (argv[0]);
    }
    break;
  default:
    usage (argv[0]);
  }

  // Initialize.
  std::vector<Row> smallTable (numOpt);
  {
    const Cell emptycell;
    const std::vector<Cell> longRow (numVariant, emptycell);
    const std::vector<Cell> shortRow (numCVariant, emptycell);
    for (uint8_t i=0; i<=maxCXXopt; ++i)
      smallTable.at(i) = longRow;
    for (uint8_t i=maxCXXopt+1; i<numOpt; ++i)
      smallTable.at(i) = shortRow;
  }
  std::vector<Row> bigTable (smallTable);

  readData (smallTable, bigTable);
  std::for_each (smallTable.begin(), smallTable.end(), (void(*)(Row&))calcStats);
  std::for_each (bigTable.begin(), bigTable.end(), (void(*)(Row&))calcStats);

  if (dump) {

    fprintf (stderr, "row = %s\n", opts.at(dumpRow).c_str());
    fprintf (stderr, "col = %s\n", variants.at(dumpCol).c_str());
    if (dumpEntireCell) {
      dumpDataGroup (bigTable.at(dumpRow).at(dumpCol).cell);
    } else {
      fprintf (stderr, "run = %u\n", dumpRun);
      dumpDataGroup (bigTable.at(dumpRow).at(dumpCol).serieses.at(dumpRun));
    }

  } else {

    printf ("%s\n", header);

    printf ("\\begin{landscape}\n");
    printf ("\\section*{Cell means with expanded uncertainty (k=%.2f)}\n\n", k);
    printf ("\\subsection*{Small}\n\n");
    printSplitTable (smallTable, cellMeanWithUncertainty);
    printf ("\\clearpage\n\\subsection*{Big}\n\n");
    printSplitTable (bigTable, cellMeanWithUncertainty);

    printf ("\\clearpage\n\\section*{Unrounded experimental standard deviation of the mean of series means}\n\n");
    printf ("\\subsection*{Small}\n\n");
    printSplitTable (smallTable, standardDeviationOfMeanOfSeriesMeans);
    printf ("\\subsection*{Big}\n\n");
    printSplitTable (bigTable, standardDeviationOfMeanOfSeriesMeans);
    printf ("\\end{landscape}\n");

    printf ("\\clearpage\n\\section*{Ratio of the experimental standard deviation of the mean of series means to the experimental standard deviation of the mean of cell data}\n\n");
    printf ("\\subsection*{Small}\n\n");
    printTable (smallTable, cellUncertaintyRatio);
    printf ("\\subsection*{Big}\n\n");
    printTable (bigTable, cellUncertaintyRatio);

    printf ("\\clearpage\n\\section*{Durbin-Watson statistics}\n\n");
    printf ("\\subsection*{Small, cell Durbin-Watson statistics}\n\n");
    printTable (smallTable, cellDurbinWatson);
    printf ("\\subsection*{Small, mean series Durbin-Watson statistics}\n\n");
    printTable (smallTable, meanSeriesDurbinWatson);
    printf ("\\subsection*{Big, cell Durbin-Watson statistics}\n\n");
    printTable (bigTable, cellDurbinWatson);
    printf ("\\subsection*{Big, mean series Durbin-Watson statistics}\n\n");
    printTable (bigTable, meanSeriesDurbinWatson);

    printf ("\\clearpage\n\\section*{Ratio of cell means, big/small}\n\n");
    printRatiosTable (smallTable, bigTable);

    // And then there's this monster.
    hugeTable (bigTable);

    printf ("\\end{document}\n");

  }
  return EXIT_SUCCESS;
}
